/*
* Copyright 2009, KyTea Development Team
* 
* Licensed under the Apache License, Version 2.0 (the "License");
* you may not use this file except in compliance with the License.
* You may obtain a copy of the License at
* 
*     http://www.apache.org/licenses/LICENSE-2.0
* 
* Unless required by applicable law or agreed to in writing, software
* distributed under the License is distributed on an "AS IS" BASIS,
* WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
* See the License for the specific language governing permissions and
* limitations under the License.
*/

#ifndef KYTEA_H__
#define KYTEA_H__

#include "kytea-config.h"
#include "kytea-struct.h"
#include "kytea-model.h"
#include "dictionary.h"

namespace kytea  {

// a class representing the main analyzer
class Kytea {

private:
    typedef unsigned FeatureId;
    typedef list<KyteaSentence*> Sentences;
    typedef vector< vector< FeatureId > > SentenceFeatures;

    StringUtil* util_;
    KyteaConfig* config_;
    Dictionary* dict_;
    Sentences sentences_;

    KyteaModel* wsModel_;

    void buildVocabulary();
    
    void trainWS();
    void prepareDictPrefixes(vector<unsigned> & dictFeats);
    void prepareNgramPrefixes(vector<KyteaString> & charPrefixes, vector<KyteaString> & typePrefixes);
    unsigned wsDictionaryFeatures(const KyteaString & sent, SentenceFeatures & feat, const vector<unsigned> & featNums);
    unsigned wsNgramFeatures(const KyteaString & sent, SentenceFeatures & feat, const vector<KyteaString> & prefixes, int n);

    void trainPE();
    unsigned peCharFeatures(const KyteaString & chars, vector<unsigned> & feat, const vector<KyteaString> & prefixes, KyteaModel * model, int n, int sc, int ec);

    void writeModels();
    // void writeModel(ofstream & out, const KyteaModel & mod);

    void readModels();
    void analyzeInput();
    void calculateWS(KyteaSentence * sent, const vector<unsigned> & dictFeats, const vector<KyteaString> & charPrefixes, const vector<KyteaString> & typePrefixes);
    
    pair<KyteaString,double> calculateUnknownPE(const KyteaString & str);
    void calculatePE(KyteaSentence * sent, const vector<KyteaString> & charPrefixes, const vector<KyteaString> & typePrefixes);

public:

    Kytea(KyteaConfig & config) : dict_(0), wsModel_(0) {
        config_ = &config;
        util_ = config.getStringUtil();
        dict_ = new Dictionary(util_);
    }
    
    ~Kytea() {
        if(dict_) delete dict_;
        if(wsModel_) delete wsModel_;
        for(Sentences::iterator it = sentences_.begin(); it != sentences_.end(); it++)
            delete *it;
    }

    // train the analyzer
    void trainAll() {
        
        // sanity check
        if(config_->getCorpusFiles().size() == 0)
            throw runtime_error("At least one input corpus must be specified (-part/-full/-prob)");
        else if(config_->getModelFile().length() == 0)
            throw runtime_error("An output model file must be specified when training (-model)");
        else if(!config_->getDoWS() && !config_->getDoPE())
            throw runtime_error("Both WS and PE are disabled. I can't build a model that estimates nothing!");
        
        // load the vocabulary, pronunciations
        buildVocabulary();

        // train the word segmenter
        if(config_->getDoWS())
            trainWS();
        
        // train the pronunciation estimator
        if(config_->getDoPE())
            trainPE();

        // write the models out to a file
        writeModels();

    }

    void analyze() {
        
        // on full input, disable word segmentation
        if(config_->getInputFormat() == CorpusIO::FORMAT_FULL)
            config_->setDoWS(false);

        // sanity check
        ostringstream buff;
        if(config_->getModelFile().length() == 0)
            throw runtime_error("A model file must be specified to run Kytea (-model)");
        else if(config_->getDoWS() == false && config_->getInputFormat() == CorpusIO::FORMAT_RAW) {
            buff << "In order to handle raw corpus input, word segmentation must be turned on." << endl
                 << "Either specify -input {full,part,prob}, stop using -nows, or train a new " << endl
                 << "model that has word segmentation included." << endl;
            throw runtime_error(buff.str());
        }
        if(!config_->getDoWS() && !config_->getDoPE())
            cerr << "Both word segmentation and pronunciation estimation are disabled!" <<endl;
        
        // read the models in from the model file
        readModels();

        // parse the sentences one by one
        analyzeInput();

    }

};

}

#endif
